"""Request models."""

from xrpl.models.auth_account import AuthAccount
from xrpl.models.path import PathStep
from xrpl.models.requests.account_channels import AccountChannels
from xrpl.models.requests.account_currencies import AccountCurrencies
from xrpl.models.requests.account_info import AccountInfo
from xrpl.models.requests.account_lines import AccountLines
from xrpl.models.requests.account_nfts import AccountNFTs
from xrpl.models.requests.account_objects import AccountObjects, AccountObjectType
from xrpl.models.requests.account_offers import AccountOffers
from xrpl.models.requests.account_tx import AccountTx
from xrpl.models.requests.amm_info import AMMInfo
from xrpl.models.requests.book_offers import BookOffers
from xrpl.models.requests.channel_authorize import ChannelAuthorize
from xrpl.models.requests.channel_verify import ChannelVerify
from xrpl.models.requests.deposit_authorized import DepositAuthorized
from xrpl.models.requests.feature import Feature
from xrpl.models.requests.fee import Fee
from xrpl.models.requests.gateway_balances import GatewayBalances
from xrpl.models.requests.generic_request import GenericRequest
from xrpl.models.requests.get_aggregate_price import GetAggregatePrice
from xrpl.models.requests.ledger import Ledger
from xrpl.models.requests.ledger_closed import LedgerClosed
from xrpl.models.requests.ledger_current import LedgerCurrent
from xrpl.models.requests.ledger_data import LedgerData
from xrpl.models.requests.ledger_entry import LedgerEntry, LedgerEntryType
from xrpl.models.requests.manifest import Manifest
from xrpl.models.requests.nft_buy_offers import NFTBuyOffers
from xrpl.models.requests.nft_history import NFTHistory
from xrpl.models.requests.nft_info import NFTInfo
from xrpl.models.requests.nft_sell_offers import NFTSellOffers
from xrpl.models.requests.nfts_by_issuer import NFTsByIssuer
from xrpl.models.requests.no_ripple_check import NoRippleCheck, NoRippleCheckRole
from xrpl.models.requests.path_find import PathFind, PathFindSubcommand
from xrpl.models.requests.ping import Ping
from xrpl.models.requests.random import Random
from xrpl.models.requests.request import Request
from xrpl.models.requests.ripple_path_find import RipplePathFind
from xrpl.models.requests.server_definitions import ServerDefinitions
from xrpl.models.requests.server_info import ServerInfo
from xrpl.models.requests.server_state import ServerState
from xrpl.models.requests.sign import Sign
from xrpl.models.requests.sign_and_submit import SignAndSubmit
from xrpl.models.requests.sign_for import SignFor
from xrpl.models.requests.simulate import Simulate
from xrpl.models.requests.submit import Submit
from xrpl.models.requests.submit_multisigned import SubmitMultisigned
from xrpl.models.requests.submit_only import SubmitOnly
from xrpl.models.requests.subscribe import StreamParameter, Subscribe, SubscribeBook
from xrpl.models.requests.transaction_entry import TransactionEntry
from xrpl.models.requests.tx import Tx
from xrpl.models.requests.unsubscribe import Unsubscribe

__all__ = [
    "AccountChannels",
    "AccountCurrencies",
    "AccountInfo",
    "AccountLines",
    "AccountNFTs",
    "AccountObjects",
    "AccountObjectType",
    "AccountOffers",
    "AccountTx",
    "AMMInfo",
    "AuthAccount",
    "BookOffers",
    "ChannelAuthorize",
    "ChannelVerify",
    "DepositAuthorized",
    "Feature",
    "Fee",
    "GatewayBalances",
    "GenericRequest",
    "GetAggregatePrice",
    "Ledger",
    "LedgerClosed",
    "LedgerCurrent",
    "LedgerData",
    "LedgerEntry",
    "LedgerEntryType",
    "Manifest",
    "NFTBuyOffers",
    "NFTSellOffers",
    "NFTInfo",
    "NFTHistory",
    "NFTsByIssuer",
    "NoRippleCheck",
    "NoRippleCheckRole",
    "PathFind",
    "PathFindSubcommand",
    "PathStep",
    "Ping",
    "Random",
    "Request",
    "RipplePathFind",
    "ServerDefinitions",
    "ServerInfo",
    "ServerState",
    "Sign",
    "SignAndSubmit",
    "SignFor",
    "Simulate",
    "Submit",
    "SubmitMultisigned",
    "SubmitOnly",
    "StreamParameter",
    "Subscribe",
    "SubscribeBook",
    "TransactionEntry",
    "Tx",
    "Unsubscribe",
]



"""Model for NFTokenCancelOffer transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Optional

from typing_extensions import Self

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class NFTokenCancelOffer(Transaction):
    """
    The NFTokenCancelOffer transaction deletes existing NFTokenOffer objects.
    It is useful if you want to free up space on your account to lower your
    reserve requirement.

    The transaction can be executed by the account that originally created
    the NFTokenOffer, the account in the `Recipient` field of the NFTokenOffer
    (if present), or any account if the NFTokenOffer has an `Expiration` and
    the NFTokenOffer has already expired.
    """

    nftoken_offers: List[str] = REQUIRED  # type: ignore
    """
    An array of identifiers of NFTokenOffer objects that should be cancelled
    by this transaction.

    It is an error if an entry in this list points to an
    object that is not an NFTokenOffer object. It is not an
    error if an entry in this list points to an object that
    does not exist. This field is required.

    :meta hide-value:
    """

    transaction_type: TransactionType = field(
        default=TransactionType.NFTOKEN_CANCEL_OFFER,
        init=False,
    )

    def _get_errors(self: Self) -> Dict[str, str]:
        return {
            key: value
            for key, value in {
                **super()._get_errors(),
                "nftoken_offers": self._get_nftoken_offers_error(),
            }.items()
            if value is not None
        }

    def _get_nftoken_offers_error(self: Self) -> Optional[str]:
        if len(self.nftoken_offers) < 1:
            return "Must specify at least one NFTokenOffer to cancel"
        return None



"""
The submit method applies a transaction and sends it to the network to be confirmed and
included in future ledgers.

This command has two modes:
* Submit-only mode takes a signed, serialized transaction as a binary blob, and submits
it to the network as-is. Since signed transaction objects are immutable, no part of the
transaction can be modified or automatically filled in after submission.
* Sign-and-submit mode takes a JSON-formatted Transaction object, completes and signs
the transaction in the same manner as the sign method, and then submits the signed
transaction. We recommend only using this mode for testing and development.

To send a transaction as robustly as possible, you should construct and sign it in
advance, persist it somewhere that you can access even after a power outage, then
submit it as a tx_blob. After submission, monitor the network with the tx method
command to see if the transaction was successfully applied; if a restart or other
problem occurs, you can safely re-submit the tx_blob transaction: it won't be applied
twice since it has the same sequence number as the old transaction.

`See submit <https://xrpl.org/submit.html>`_
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Optional, Type

from typing_extensions import Self

from xrpl.constants import CryptoAlgorithm
from xrpl.models.requests.submit import Submit
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class SignAndSubmit(Submit):
    """
    The submit method applies a transaction and sends it to the network to be confirmed
    and included in future ledgers.

    This command has two modes:
    * Submit-only mode takes a signed, serialized transaction as a binary blob, and
    submits it to the network as-is. Since signed transaction objects are immutable, no
    part of the transaction can be modified or automatically filled in after submission.
    * Sign-and-submit mode takes a JSON-formatted Transaction object, completes and
    signs the transaction in the same manner as the sign method, and then submits the
    signed transaction. We recommend only using this mode for testing and development.

    To send a transaction as robustly as possible, you should construct and sign it in
    advance, persist it somewhere that you can access even after a power outage, then
    submit it as a tx_blob. After submission, monitor the network with the tx method
    command to see if the transaction was successfully applied; if a restart or other
    problem occurs, you can safely re-submit the tx_blob transaction: it won't be
    applied twice since it has the same sequence number as the old transaction.

    `See submit <https://xrpl.org/submit.html>`_
    """

    transaction: Transaction = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    secret: Optional[str] = None
    seed: Optional[str] = None
    seed_hex: Optional[str] = None
    passphrase: Optional[str] = None
    key_type: Optional[CryptoAlgorithm] = None
    offline: bool = False
    build_path: Optional[bool] = None  # note: None does have meaning here
    fee_mult_max: int = 10
    fee_div_max: int = 1

    @classmethod
    def from_dict(cls: Type[Self], value: Dict[str, Any]) -> Self:
        """
        Construct a new SignAndSubmit from a dictionary of parameters.

        Args:
            value: The value to construct the SignAndSubmit from.

        Returns:
            A new SignAndSubmit object, constructed using the given parameters.
        """
        if "tx_json" in value:
            fixed_value = {
                **value,
                "transaction": Transaction.from_xrpl(value["tx_json"]),
            }
            del fixed_value["tx_json"]
        else:
            fixed_value = value
        return super().from_dict(fixed_value)

    def to_dict(self: Self) -> Dict[str, Any]:
        """
        Returns the dictionary representation of a SignAndSubmit.

        Returns:
            The dictionary representation of a SignAndSubmit.
        """
        return_dict = super().to_dict()
        del return_dict["transaction"]
        return_dict["tx_json"] = self.transaction.to_xrpl()
        return return_dict

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()
        if not self._has_only_one_seed():
            errors["SignAndSubmit"] = (
                "Must have only one of `secret`, `seed`, `seed_hex`, and `passphrase`."
            )

        if self.secret is not None and self.key_type is not None:
            errors["key_type"] = "Must omit `key_type` if `secret` is provided."

        return errors

    def _has_only_one_seed(self: Self) -> bool:
        present_items = [
            item
            for item in [self.secret, self.seed, self.seed_hex, self.passphrase]
            if item is not None
        ]
        return len(present_items) == 1



"""A client for interacting with the rippled WebSocket API."""

from __future__ import annotations

from collections.abc import AsyncIterator
from types import TracebackType
from typing import Any, Dict, Type

from typing_extensions import Self

from xrpl.asyncio.clients.async_client import AsyncClient
from xrpl.asyncio.clients.client import REQUEST_TIMEOUT
from xrpl.asyncio.clients.exceptions import XRPLWebsocketException
from xrpl.asyncio.clients.websocket_base import WebsocketBase
from xrpl.models.requests.request import Request
from xrpl.models.response import Response


class AsyncWebsocketClient(AsyncClient, WebsocketBase):
    """
    An async client for interacting with the rippled WebSocket API.

    Instead of calling ``open`` and ``close`` yourself, you
    can use a context like so::

        async with AsyncWebsocketClient(url) as client:
            # inside the context the client is open
        # after exiting the context, the client is closed

    Doing this will open and close the client for you and is
    preferred.

    NOTE: if you are not using subscriptions or other WebSocket-only
    features of rippled, you may not need to do anything other than
    open the client and make requests::

        from xrpl.asyncio.clients import AsyncWebsocketClient
        from xrpl.asyncio.ledger import get_fee
        from xrpl.models import Fee


        async with AsyncWebsocketClient(url) as client:
            # using helper functions
            print(await get_fee(client))

            # using raw requests yourself
            print(await client.request(Fee())

    However, if you are using some functionality that makes use of
    subscriptions or other "websocket-y" things, you can iterate over
    the client like so to read incoming messages::

        async with AsyncWebsocketClient(url) as client:
            # inside the context the client is open
            async for message in client:
                # do something with a message
        # after exiting the context, the client is closed

    The recommended way to use this client is to set up a Task
    using the ``asyncio`` library to listen to incoming
    messages and do something with them, but the above will
    work fine if you want to listen indefinitely. This is how
    you can use a Task to listen to messages without blocking. Note this
    example can be copied directly into a file and run with no changes::

        import asyncio

        from xrpl.asyncio.clients import AsyncWebsocketClient
        from xrpl.models import Subscribe, Unsubscribe, StreamParameter

        URL = "wss://s.altnet.rippletest.net:51233"


        def on_open():
            print("I have opened a connection!")


        async def on_message(client):
            async for message in client:
                # do something with a message - we'll just print
                print(message)


        def on_close():
            print("I have closed a connection cleanly!")


        def on_error():
            print("An error occurred!")


        async def main():
            # we'll use this to determine if the websocket closed cleanly or
            # not
            error_happened = False

            async with AsyncWebsocketClient(URL) as client:
                try:
                    # here you'll run any code that should happen immediately
                    # after the connection is made. this is equivalent to the
                    # javascript 'open' event
                    on_open()

                    # set up the `on_message` function as a Task
                    # so that it doesn't wait for a response, but
                    # will "awaken" whenever the `asyncio` event
                    # loop toggles to it. this is equivalent to the javascript
                    # 'message' event
                    asyncio.create_task(on_message(client))

                    # now, the `on_message` function will run as if
                    # it were "in the background", doing whatever you
                    # want as soon as it has a message.

                    # now let's subscribe to something. in this case,
                    # we can just use `send` instead of `request`
                    # because we don't really care about the response
                    # since the `on_message` handler will also get it.
                    await client.send(Subscribe(
                        streams=[StreamParameter.LEDGER],
                    ))
                    print("Subscribed to the ledger!")

                    # in the meantime, you can continue to do whatever
                    # you want and the python `asyncio` event loop
                    # will toggle between your code and the listener
                    # as messages are ready. let's just sleep. note,
                    # you need to use `asyncio.sleep` within
                    # async code instead of `time.sleep`, otherwise
                    # you will block all the waiting tasks instead of
                    # just this code path.
                    await asyncio.sleep(50)

                    # now that we're done, we can unsubscribe if
                    # we like
                    await client.send(Unsubscribe(
                        streams=[StreamParameter.LEDGER],
                    ))
                    print("Unsubscribed from the ledger!")
                except:
                    # if you wish you perform some logic when the websocket
                    # connection closes due to error, you can catch and run
                    # whatever you need to here. this is equivalent to the
                    # javascript 'error' event
                    error_happened = True
                    on_error()
            # now, outside of the context, the client is closed.
            # the `on_message` task will now never receive a new message. you
            # can now run any code you need to run after the connection is
            # closed. this is equivalent to the javascript 'close' event
            if not error_happened:
                on_close()


        if __name__ == "__main__":
            # remember to run your entire program within a
            # `asyncio.run` call.
            asyncio.run(main())


    If you need to ensure that you reconnect your websockets whenever they
    disconnect, you can create a supervisor like the example below. Note this
    example can be copied and run directly with no changes::

        import asyncio

        from xrpl.asyncio.clients import AsyncWebsocketClient
        from xrpl.models import Subscribe, StreamParameter

        URL = "wss://s.altnet.rippletest.net:51233"


        async def websocket_supervisor():
            # whenever the websocket disconnects, for any reason, the loop
            # will restart, reconnect, and set everything up again.
            while True:
                try:
                    await long_websocket_task()
                except:
                    print("Lost connection! Reconnecting")


        async def on_message(client):
            async for message in client:
                print(message)


        async def long_websocket_task():
            async with AsyncWebsocketClient(URL) as client:
                # set up a listener task
                listener = asyncio.create_task(on_message(client))

                # subscribe to the ledger
                await client.send(Subscribe(
                    streams=[StreamParameter.LEDGER],
                ))

                # sleep infinitely until the connection closes on us
                while client.is_open():
                    await asyncio.sleep(0)
                listener.cancel()


        async def main():
            await websocket_supervisor()


        if __name__ == "__main__":
            asyncio.run(main())
    """

    async def open(self: Self) -> None:
        """Connects the client to the Web Socket API at the given URL."""
        if not self.is_open():
            await self._do_open()

    async def close(self: Self) -> None:
        """Closes the connection."""
        if self.is_open():
            await self._do_close()

    async def __aenter__(self: Self) -> Self:
        """
        Enters an async context after opening itself.

        Returns:
            The opened client.
        """
        await self.open()
        return self

    async def __aexit__(
        self: Self,
        _exc_type: Type[BaseException],
        _exc_val: BaseException,
        _trace: TracebackType,
    ) -> None:
        """Exits an async context after closing itself."""
        await self.close()

    async def __aiter__(self: Self) -> AsyncIterator[Dict[str, Any]]:
        """
        Iterate on received messages.

        Yields:
            Message at the top of the queue.
        """
        while self.is_open():
            yield await self._do_pop_message()

    async def send(self: Self, request: Request) -> None:
        """
        Submit the request represented by the request to the
        rippled node specified by this client's URL. Unlike ``request``,
        ``send`` does not wait for this request's response. In many cases
        it may be more convenient to use ``request``.

        Arguments:
            request: A Request object representing information about a rippled request.

        Raises:
            XRPLWebsocketException: If there is already an open request by the
                request's ID, or if this WebsocketBase is not open.
        """
        if not self.is_open():
            raise XRPLWebsocketException("Websocket is not open")
        await self._do_send(request)

    async def _request_impl(
        self: Self, request: Request, *, timeout: float = REQUEST_TIMEOUT
    ) -> Response:
        """
        ``_request_impl`` implementation for async websocket.

        Arguments:
            request: An object representing information about a rippled request.

        Returns:
            The response from the server, as a Response object.

        Raises:
            XRPLWebsocketException: If there is already an open request by the
                request's ID, or if this WebsocketBase is not open.

        :meta private:
        """
        if not self.is_open():
            raise XRPLWebsocketException("Websocket is not open")
        return await self._do_request_impl(request, timeout)



"""Context manager and helpers for the serialization of a JSON object into bytes."""

from __future__ import annotations  # Requires Python 3.7+

from typing_extensions import Final, Self

from xrpl.core.binarycodec.definitions.field_instance import FieldInstance
from xrpl.core.binarycodec.types.serialized_type import SerializedType

# Constants used in length prefix encoding:
# max length that can be represented in a single byte per XRPL serialization encoding
_MAX_SINGLE_BYTE_LENGTH: Final[int] = 192
# max length that can be represented in 2 bytes per XRPL serialization encoding
_MAX_DOUBLE_BYTE_LENGTH: Final[int] = 12481
# max value that can be used in the second byte of a length field
_MAX_SECOND_BYTE_VALUE: Final[int] = 240
# maximum length that can be encoded in a length prefix per XRPL serialization encoding
_MAX_LENGTH_VALUE: Final[int] = 918744


def _encode_variable_length_prefix(length: int) -> bytes:
    """
    Helper function for length-prefixed fields including Blob types
    and some AccountID types. Calculates the prefix of variable length bytes.

    The length of the prefix is 1-3 bytes depending on the length of the contents:
    Content length <= 192 bytes: prefix is 1 byte
    192 bytes < Content length <= 12480 bytes: prefix is 2 bytes
    12480 bytes < Content length <= 918744 bytes: prefix is 3 bytes

    `See Length Prefixing <https://xrpl.org/serialization.html#length-prefixing>`_
    """
    if length <= _MAX_SINGLE_BYTE_LENGTH:
        return length.to_bytes(1, byteorder="big", signed=False)
    if length < _MAX_DOUBLE_BYTE_LENGTH:
        length -= _MAX_SINGLE_BYTE_LENGTH + 1
        byte1 = ((length >> 8) + (_MAX_SINGLE_BYTE_LENGTH + 1)).to_bytes(
            1, byteorder="big", signed=False
        )
        byte2 = (length & 0xFF).to_bytes(1, byteorder="big", signed=False)
        return byte1 + byte2
    if length <= _MAX_LENGTH_VALUE:
        length -= _MAX_DOUBLE_BYTE_LENGTH
        byte1 = ((_MAX_SECOND_BYTE_VALUE + 1) + (length >> 16)).to_bytes(
            1, byteorder="big", signed=False
        )
        byte2 = ((length >> 8) & 0xFF).to_bytes(1, byteorder="big", signed=False)
        byte3 = (length & 0xFF).to_bytes(1, byteorder="big", signed=False)
        return byte1 + byte2 + byte3

    raise ValueError(f"VariableLength field must be <= {_MAX_LENGTH_VALUE} bytes long")


class BinarySerializer:
    """Serializes JSON to XRPL binary format."""

    def __init__(self: Self) -> None:
        """Construct a BinarySerializer."""
        self.bytesink = bytes()

    def append(self: Self, bytes_object: bytes) -> None:
        """
        Write given bytes to this BinarySerializer's bytesink.

        Args:
            bytes_object: The bytes to write to bytesink.
        """
        self.bytesink += bytes_object

    def __bytes__(self: Self) -> bytes:
        """
        Get the bytes representation of a BinarySerializer.

        Returns:
            The bytes representation of the BinarySerializer's bytesink.
        """
        return self.bytesink

    def write_length_encoded(
        self: Self,
        value: SerializedType,
        encode_value: bool = True,
    ) -> None:
        """
        Write a variable length encoded value to the BinarySerializer.

        Args:
            value: The SerializedType object to write to bytesink.
            encode_value: Does not encode the value; just encodes `00` in its place.
                Used in the UNLModify encoding workaround. The default is True.
        """
        byte_object = bytearray()
        if encode_value:
            value.to_byte_sink(byte_object)
        length_prefix = _encode_variable_length_prefix(len(byte_object))
        self.bytesink += length_prefix
        self.bytesink += byte_object

    def write_field_and_value(
        self: Self,
        field: FieldInstance,
        value: SerializedType,
        is_unl_modify_workaround: bool = False,
    ) -> None:
        """
        Write field and value to the buffer.

        Args:
            field: The field to write to the buffer.
            value: The value to write to the buffer.
            is_unl_modify_workaround: Encode differently for UNLModify
                pseudotransactions, due to a bug in rippled. Only True for the Account
                field in UNLModify pseudotransactions. The default is False.
        """
        self.bytesink += bytes(field.header)

        if field.is_variable_length_encoded:
            self.write_length_encoded(value, not is_unl_modify_workaround)
        else:
            self.bytesink += bytes(value)



"""
The book_offers method retrieves a list of offers, also known
as the order book, between two currencies.
"""

from dataclasses import dataclass, field
from typing import Optional

from xrpl.models.currencies import Currency
from xrpl.models.requests.request import LookupByLedgerRequest, Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class BookOffers(Request, LookupByLedgerRequest):
    """
    The book_offers method retrieves a list of offers, also known
    as the order book, between two currencies.
    """

    taker_gets: Currency = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    taker_pays: Currency = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    method: RequestMethod = field(default=RequestMethod.BOOK_OFFERS, init=False)
    limit: Optional[int] = None
    taker: Optional[str] = None



"""
Codec for serializing and deserializing blob fields.
See `Blob Fields <https://xrpl.org/serialization.html#blob-fields>`_
"""

from __future__ import annotations

from typing import Type

from typing_extensions import Self

from xrpl.core.binarycodec.binary_wrappers.binary_parser import BinaryParser
from xrpl.core.binarycodec.exceptions import XRPLBinaryCodecException
from xrpl.core.binarycodec.types.serialized_type import SerializedType


class Blob(SerializedType):
    """
    Codec for serializing and deserializing blob fields.
    See `Blob Fields <https://xrpl.org/serialization.html#blob-fields>`_
    """

    def __init__(self: Self, buffer: bytes) -> None:
        """Construct a new Blob type from a ``bytes`` value."""
        super().__init__(buffer)

    @classmethod
    def from_parser(cls: Type[Self], parser: BinaryParser, length_hint: int) -> Self:
        """
        Defines how to read a Blob from a BinaryParser.

        Args:
            parser: The parser to construct a Blob from.
            length_hint: The number of bytes to consume from the parser.

        Returns:
            The Blob constructed from parser.
        """
        return cls(parser.read(length_hint))

    @classmethod
    def from_value(cls: Type[Self], value: str) -> Self:
        """
        Create a Blob object from a hex-string.

        Args:
            value: The hex-encoded string to construct a Blob from.

        Returns:
            The Blob constructed from value.

        Raises:
            XRPLBinaryCodecException: If the Blob can't be constructed from value.
        """
        if not isinstance(value, str):
            raise XRPLBinaryCodecException(
                "Invalid type to construct a Blob: expected str, received "
                f"{value.__class__.__name__}."
            )

        if isinstance(value, str):
            return cls(bytes.fromhex(value))

        raise XRPLBinaryCodecException("Cannot construct Blob from value given")



"""Model for a XChainAddClaimAttestation transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, Union

from typing_extensions import Literal

from xrpl.models.amounts import Amount
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init
from xrpl.models.xchain_bridge import XChainBridge


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class XChainAddClaimAttestation(Transaction):
    """
    Represents a XChainAddClaimAttestation transaction.
    The ``XChainAddClaimAttestation`` transaction provides proof from a witness
    server, attesting to an ``XChainCommit`` transaction.
    """

    xchain_bridge: XChainBridge = REQUIRED  # type: ignore
    """
    The bridge to use to transfer funds. This field is required.

    :meta hide-value:
    """

    public_key: str = REQUIRED  # type: ignore
    """
    The public key used to verify the signature. This field is required.

    :meta hide-value:
    """

    signature: str = REQUIRED  # type: ignore
    """
    The signature attesting to the event on the other chain. This field is
    required.

    :meta hide-value:
    """

    other_chain_source: str = REQUIRED  # type: ignore
    """
    The account on the source chain that submitted the ``XChainCommit``
    transaction that triggered the event associated with the attestation. This
    field is required.

    :meta hide-value:
    """

    amount: Amount = REQUIRED  # type: ignore
    """
    The amount committed by the ``XChainCommit`` transaction on the source
    chain. This field is required.

    :meta hide-value:
    """

    attestation_reward_account: str = REQUIRED  # type: ignore
    """
    The account that should receive this signer's share of the
    ``SignatureReward``. This field is required.

    :meta hide-value:
    """

    attestation_signer_account: str = REQUIRED  # type: ignore
    """
    The account on the door account's signer list that is signing the
    transaction. This field is required.

    :meta hide-value:
    """

    was_locking_chain_send: Union[Literal[0], Literal[1]] = REQUIRED  # type: ignore
    """
    A boolean representing the chain where the event occurred. This field is
    required.

    :meta hide-value:
    """

    xchain_claim_id: Union[str, int] = REQUIRED  # type: ignore
    """
    The ``XChainClaimID`` associated with the transfer, which was included in
    the ``XChainCommit`` transaction. This field is required.

    :meta hide-value:
    """

    destination: Optional[str] = None
    """
    The destination account for the funds on the destination chain (taken from
    the ``XChainCommit`` transaction).

    :meta hide-value:
    """

    transaction_type: TransactionType = field(
        default=TransactionType.XCHAIN_ADD_CLAIM_ATTESTATION,
        init=False,
    )



"""
The `nft_history` method retreives a list of transactions that involved the
specified NFToken.
"""

from dataclasses import dataclass, field
from typing import Any, Optional

from xrpl.models.requests.request import LookupByLedgerRequest, Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class NFTHistory(Request, LookupByLedgerRequest):
    """
    The `nft_history` method retreives a list of transactions that involved the
    specified NFToken.
    """

    method: RequestMethod = field(default=RequestMethod.NFT_HISTORY, init=False)
    nft_id: str = REQUIRED  # type: ignore
    """
    The unique identifier of an NFToken.
    The request returns past transactions of this NFToken. This value is required.

    :meta hide-value:
    """

    ledger_index_min: Optional[int] = None
    ledger_index_max: Optional[int] = None
    binary: bool = False
    forward: bool = False
    limit: Optional[int] = None
    # marker data shape is actually undefined in the spec, up to the
    # implementation of an individual server
    marker: Optional[Any] = None



"""Model for CredentialCreate transaction type."""

from dataclasses import dataclass, field
from typing import Dict, Optional

from typing_extensions import Self

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import (
    HEX_REGEX,
    KW_ONLY_DATACLASS,
    get_credential_type_error,
    require_kwargs_on_init,
)

_MAX_URI_LENGTH = 256


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class CredentialCreate(Transaction):
    """This transaction creates a Credential object. It must be sent by the issuer."""

    account: str = REQUIRED  # type: ignore
    """
    The issuer of the credential.
    """

    subject: str = REQUIRED  # type: ignore
    """
    The subject of the credential.
    """

    credential_type: str = REQUIRED  # type: ignore
    """
    A hex-encoded value to identify the type of credential from the issuer.
    """

    expiration: Optional[int] = None
    """
    The credential expiration.
    """

    uri: Optional[str] = None
    """
    Additional data about the credential (such as a link to the Verifiable
    Credential document).
    """

    transaction_type: TransactionType = field(
        default=TransactionType.CREDENTIAL_CREATE,
        init=False,
    )

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()

        if (uri_error := self._get_uri_error()) is not None:
            errors["uri"] = uri_error

        if (
            cred_type_error := get_credential_type_error(self.credential_type)
        ) is not None:
            errors["credential_type"] = cred_type_error

        return errors

    def _get_uri_error(self: Self) -> Optional[str]:
        if self.uri is None:
            return None

        errors = []
        if len(self.uri) == 0:
            errors.append("cannot be an empty string.")
        elif len(self.uri) > _MAX_URI_LENGTH:
            errors.append(f"Length cannot exceed {_MAX_URI_LENGTH} characters.")

        if not HEX_REGEX.fullmatch(self.uri):
            errors.append("Must be encoded in hex.")
        return " ".join(errors) if errors else None



"""High-level methods to obtain information about account transaction history."""

import asyncio

from xrpl.asyncio.account import transaction_history
from xrpl.clients.sync_client import SyncClient
from xrpl.models.response import Response


def get_latest_transaction(account: str, client: SyncClient) -> Response:
    """
    Fetches the most recent transaction on the ledger associated with an account.

    Args:
        account: the account to query.
        client: the network client used to communicate with a rippled node.

    Returns:
        The Response object containing the transaction info.

    Raises:
        XRPLRequestFailureException: if the transaction fails.
    """
    return asyncio.run(transaction_history.get_latest_transaction(account, client))



"""
An Amount is an object specifying a currency, a quantity of that currency, and the
counterparty (issuer) on the trustline that holds the value. For XRP, there is no
counterparty.
"""

from typing import Union

from typing_extensions import TypeGuard

from xrpl.models.amounts.issued_currency_amount import IssuedCurrencyAmount
from xrpl.models.amounts.mpt_amount import MPTAmount

Amount = Union[IssuedCurrencyAmount, MPTAmount, str]


def is_xrp(amount: Amount) -> TypeGuard[str]:
    """
    Returns whether amount is an XRP value, as opposed to an issued currency
    or MPT value.

    Args:
        amount: The Amount object whose type is being checked.

    Returns:
        Whether the amount is an XRP value.
    """
    return isinstance(amount, str)


def is_issued_currency(amount: Amount) -> TypeGuard[IssuedCurrencyAmount]:
    """
    Returns whether amount is an issued currency value, as opposed to an XRP
    or MPT value.

    Args:
        amount: The Amount object whose type is being checked.

    Returns:
        Whether the amount is an issued currency value.
    """
    return isinstance(amount, IssuedCurrencyAmount)


def is_mpt(amount: Amount) -> TypeGuard[MPTAmount]:
    """
    Returns whether amount is an MPT value, as opposed to an XRP or
    an issued currency value.

    Args:
        amount: The Amount object whose type is being checked.

    Returns:
        Whether the amount is an issued currency value.
    """
    return isinstance(amount, MPTAmount)


def get_amount_value(amount: Amount) -> float:
    """
    Returns the value of an amount irrespective of its currency.

    Args:
        amount: The Amount object whose value we want.

    Returns:
        The value of the amount irrespective of its currency.
    """
    if is_xrp(amount):
        return float(amount)
    if is_issued_currency(amount):
        return float(amount.value)
    if is_mpt(amount):
        return float(amount.value)
    raise ValueError(f"Invalid amount: {repr(amount)}")



"""
An Amount is an object specifying a currency, a quantity
of that currency, and the counterparty (issuer) on the trustline
that holds the value. For XRP, there is no counterparty.
"""

from xrpl.models.amounts.amount import (
    Amount,
    get_amount_value,
    is_issued_currency,
    is_mpt,
    is_xrp,
)
from xrpl.models.amounts.issued_currency_amount import IssuedCurrencyAmount
from xrpl.models.amounts.mpt_amount import MPTAmount

__all__ = [
    "Amount",
    "IssuedCurrencyAmount",
    "is_xrp",
    "is_issued_currency",
    "is_mpt",
    "get_amount_value",
    "MPTAmount",
]



"""
Class for serializing and deserializing an 8-bit UInt.
See `UInt Fields <https://xrpl.org/serialization.html#uint-fields>`_
"""

from __future__ import annotations

from typing import Optional, Type

from typing_extensions import Final, Self

from xrpl.core.binarycodec.binary_wrappers.binary_parser import BinaryParser
from xrpl.core.binarycodec.exceptions import XRPLBinaryCodecException
from xrpl.core.binarycodec.types.uint import UInt

_WIDTH: Final[int] = 1  # 8 / 8


class UInt8(UInt):
    """
    Class for serializing and deserializing an 8-bit UInt.
    See `UInt Fields <https://xrpl.org/serialization.html#uint-fields>`_
    """

    def __init__(self: Self, buffer: bytes = bytes(_WIDTH)) -> None:
        """Construct a new UInt8 type from a ``bytes`` value."""
        super().__init__(buffer)

    @classmethod
    def from_parser(
        cls: Type[Self], parser: BinaryParser, _length_hint: Optional[int] = None
    ) -> Self:
        """
        Construct a new UInt8 type from a BinaryParser.

        Args:
            parser: The parser to construct a UInt8 from.

        Returns:
            A new UInt8.
        """
        return cls(parser.read(_WIDTH))

    @classmethod
    def from_value(cls: Type[Self], value: int) -> Self:
        """
        Construct a new UInt8 type from a number.

        Args:
            value: The value to construct a UInt8 from.

        Returns:
            A new UInt8.

        Raises:
            XRPLBinaryCodecException: If a UInt8 cannot be constructed.
        """
        if not isinstance(value, int):
            raise XRPLBinaryCodecException(
                "Invalid type to construct a UInt8: expected int, "
                f"received {value.__class__.__name__}."
            )

        if isinstance(value, int):
            value_bytes = (value).to_bytes(_WIDTH, byteorder="big", signed=False)
            return cls(value_bytes)

        raise XRPLBinaryCodecException("Cannot construct UInt8 from given value")



"""
The manifest method reports the current
"manifest" information for a given validator
public key. The "manifest" is the public portion
of that validator's configured token.
"""

from dataclasses import dataclass, field

from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class Manifest(Request):
    """
    The manifest method reports the current
    "manifest" information for a given validator
    public key. The "manifest" is the public portion
    of that validator's configured token.
    """

    method: RequestMethod = field(default=RequestMethod.MANIFEST, init=False)
    public_key: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """



"""This module contains commonly-used constants."""

from typing_extensions import Final

# The dictionary used for XRPL base58 encodings
XRPL_ALPHABET: Final[bytes] = (
    b"rpshnaf39wBUDNEGHJKLM4PQRST7VWXYZ2bcdeCg65jkm8oFqi1tuvAxyz"
)



"""Model for a XChainClaim transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Optional, Union

from typing_extensions import Self

from xrpl.models.amounts import Amount, is_issued_currency, is_mpt, is_xrp
from xrpl.models.currencies import XRP, Currency
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init
from xrpl.models.xchain_bridge import XChainBridge


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class XChainClaim(Transaction):
    """
    Represents a XChainClaim transaction.
    The ``XChainClaim`` transaction completes a cross-chain transfer of value.
    It allows a user to claim the value on the destination chain - the
    equivalent of the value locked on the source chain.
    """

    xchain_bridge: XChainBridge = REQUIRED  # type: ignore
    """
    The bridge to use for the transfer. This field is required.

    :meta hide-value:
    """

    xchain_claim_id: Union[int, str] = REQUIRED  # type: ignore
    """
    The unique integer ID for the cross-chain transfer that was referenced in
    the corresponding ``XChainCommit`` transaction. This field is required.

    :meta hide-value:
    """

    destination: str = REQUIRED  # type: ignore
    """
    The destination account on the destination chain. It must exist or the
    transaction will fail. However, if the transaction fails in this case, the
    sequence number and collected signatures won't be destroyed, and the
    transaction can be rerun with a different destination. This field is
    required.

    :meta hide-value:
    """

    destination_tag: Optional[int] = None
    """
    An integer destination tag.

    :meta hide-value:
    """

    amount: Amount = REQUIRED  # type: ignore
    """
    The amount to claim on the destination chain. This must match the amount
    attested to on the attestations associated with this ``XChainClaimID``.
    This field is required.

    :meta hide-value:
    """

    transaction_type: TransactionType = field(
        default=TransactionType.XCHAIN_CLAIM,
        init=False,
    )

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()

        bridge = self.xchain_bridge

        amount = self.amount
        if is_xrp(amount):
            currency: Currency = XRP()
        elif is_issued_currency(amount):
            currency = amount.to_currency()
        elif is_mpt(amount):
            currency = amount.to_currency()
        else:
            errors["amount"] = "Currency can't be derived."

        if (
            currency != bridge.locking_chain_issue
            and currency != bridge.issuing_chain_issue
        ):
            errors["amount"] = (
                "amount must match either locking chain issue or issuing chain issue."
            )

        return errors



"""Model for CredentialAccept transaction type."""

from dataclasses import dataclass, field
from typing import Dict

from typing_extensions import Self

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import (
    KW_ONLY_DATACLASS,
    get_credential_type_error,
    require_kwargs_on_init,
)


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class CredentialAccept(Transaction):
    """This transaction accepts a credential issued to the Account (i.e. the Account is
    the Subject of the Credential object). The credential is not considered valid until
    it has been transferred/accepted.
    """

    account: str = REQUIRED  # type: ignore
    """
    The subject of the credential.
    """

    issuer: str = REQUIRED  # type: ignore
    """
    The issuer of the credential.
    """

    credential_type: str = REQUIRED  # type: ignore
    """
    A hex-encoded value to identify the type of credential from the issuer.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.CREDENTIAL_ACCEPT,
        init=False,
    )

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()
        if (
            cred_type_error := get_credential_type_error(self.credential_type)
        ) is not None:
            errors["credential_type"] = cred_type_error
        return errors



"""Model for OfferCreate transaction type."""

from dataclasses import dataclass, field
from enum import Enum
from typing import Optional

from xrpl.models.amounts import Amount
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction, TransactionFlagInterface
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


class OfferCreateFlag(int, Enum):
    """
    Transactions of the OfferCreate type support additional values in the Flags field.
    This enum represents those options.

    `See OfferCreate Flags <https://xrpl.org/offercreate.html#offercreate-flags>`_
    """

    TF_PASSIVE = 0x00010000
    """
    If enabled, the offer does not consume offers that exactly match it, and instead
    becomes an Offer object in the ledger. It still consumes offers that cross it.
    """

    TF_IMMEDIATE_OR_CANCEL = 0x00020000
    """
    Treat the offer as an `Immediate or Cancel order
    <https://en.wikipedia.org/wiki/Immediate_or_cancel>`_. If enabled, the offer
    never becomes a ledger object: it only tries to match existing offers in the
    ledger. If the offer cannot match any offers immediately, it executes
    "successfully" without trading any currency. In this case, the transaction has
    the result code `tesSUCCESS`, but creates no Offer objects in the ledger.
    """

    TF_FILL_OR_KILL = 0x00040000
    """
    Treat the offer as a `Fill or Kill order
    <https://en.wikipedia.org/wiki/Fill_or_kill>`_. Only try to match existing
    offers in the ledger, and only do so if the entire `TakerPays` quantity can be
    obtained. If the `fix1578 amendment
    <https://xrpl.org/known-amendments.html#fix1578>`_ is enabled and the offer
    cannot be executed when placed, the transaction has the result code `tecKILLED`;
    otherwise, the transaction uses the result code `tesSUCCESS` even when it was
    killed without trading any currency.
    """

    TF_SELL = 0x00080000
    """
    Exchange the entire `TakerGets` amount, even if it means obtaining more than the
    `TakerPays amount` in exchange.
    """


class OfferCreateFlagInterface(TransactionFlagInterface):
    """
    Transactions of the OfferCreate type support additional values in the Flags field.
    This TypedDict represents those options.

    `See OfferCreate Flags <https://xrpl.org/offercreate.html#offercreate-flags>`_
    """

    TF_PASSIVE: bool
    TF_IMMEDIATE_OR_CANCEL: bool
    TF_FILL_OR_KILL: bool
    TF_SELL: bool


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class OfferCreate(Transaction):
    """
    Represents an `OfferCreate <https://xrpl.org/offercreate.html>`_ transaction,
    which executes a limit order in the `decentralized exchange
    <https://xrpl.org/decentralized-exchange.html>`_. If the specified exchange
    cannot be completely fulfilled, it creates an Offer object for the remainder.
    Offers can be partially fulfilled.
    """

    taker_gets: Amount = REQUIRED  # type: ignore
    """
    The amount and type of currency being provided by the sender of this
    transaction. This field is required.

    :meta hide-value:
    """

    taker_pays: Amount = REQUIRED  # type: ignore
    """
    The amount and type of currency the sender of this transaction wants in
    exchange for the full ``taker_gets`` amount. This field is required.

    :meta hide-value:
    """

    expiration: Optional[int] = None
    """
    Time after which the offer is no longer active, in seconds since the
    Ripple Epoch.
    """

    offer_sequence: Optional[int] = None
    """
    The Sequence number (or Ticket number) of a previous OfferCreate to cancel
    when placing this Offer.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.OFFER_CREATE,
        init=False,
    )



"""Interface for all async network clients to follow."""

from __future__ import annotations

from typing_extensions import Self

from xrpl.asyncio.clients.client import Client
from xrpl.models.requests.request import Request
from xrpl.models.response import Response


class AsyncClient(Client):
    """
    Interface for all async network clients to follow.

    :meta private:
    """

    async def request(self: Self, request: Request) -> Response:
        """
        Makes a request with this client and returns the response.

        Arguments:
            request: The Request to send.

        Returns:
            The Response for the given Request.
        """
        return await self._request_impl(request)



"""The base model for all pseudo-transactions and their nested object types."""

from dataclasses import dataclass, field

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import PseudoTransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init

_ACCOUNT_ZERO = "rrrrrrrrrrrrrrrrrrrrrhoLvTp"  # base58 encoding of the value `0`


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class PseudoTransaction(Transaction):
    """
    Pseudo-transactions are never submitted by users, nor propagated through the
    network. Instead, a server may choose to inject pseudo-transactions in a proposed
    ledger directly according to specific protocol rules. If enough servers inject an
    identical pseudo-transaction for it to be approved by the consensus process, then
    the pseudo-transaction becomes included in the ledger, and appears in ledger data
    thereafter.
    """

    account: str = field(default=_ACCOUNT_ZERO, init=False)
    fee: str = field(default="0", init=False)
    sequence: int = field(default=0, init=False)
    signing_pub_key: str = field(default="", init=False)
    txn_signature: str = field(default="", init=False)
    source_tag: None = field(default=None, init=False)
    transaction_type: PseudoTransactionType = REQUIRED  # type: ignore



"""Parse balance changes of every account involved in the given transaction."""

from decimal import Decimal
from typing import List, Optional

from xrpl.models import TransactionMetadata
from xrpl.utils.txn_parser.utils import (
    AccountBalances,
    NormalizedNode,
    derive_account_balances,
    get_value,
)


def get_balance_changes(metadata: TransactionMetadata) -> List[AccountBalances]:
    """
    Parse all balance changes from a transaction's metadata. Does not include
    MPT balance.

    Args:
        metadata: Transactions metadata.

    Returns:
        All balance changes caused by a transaction.
        The balance changes are grouped by the affected account addresses.
    """
    # TODO: add support for MPT balance
    return derive_account_balances(metadata, _compute_balance_change)


def _compute_balance_change(node: NormalizedNode) -> Optional[Decimal]:
    """
    Get the balance change from a node.

    Args:
        node: The affected node.

    Returns:
        The balance change.
    """
    value: Optional[Decimal] = None
    new_fields = node.get("NewFields")
    previous_fields = node.get("PreviousFields")
    final_fields = node.get("FinalFields")
    if new_fields is not None:
        balance = new_fields.get("Balance")
        if balance is not None:
            value = get_value(balance)
    elif previous_fields is not None and final_fields is not None:
        previous_fields_balance = previous_fields.get("Balance")
        final_fields_balance = final_fields.get("Balance")
        if previous_fields_balance is not None and final_fields_balance is not None:
            value = get_value(final_fields_balance) - get_value(previous_fields_balance)
    if value is None or value == Decimal(0):
        return None
    return value



"""The base class for all binary codec field types."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Type

from typing_extensions import Self

if TYPE_CHECKING:
    # To prevent a circular dependency.
    from xrpl.core.binarycodec.binary_wrappers.binary_parser import BinaryParser


class SerializedType(ABC):
    """The base class for all binary codec field types."""

    def __init__(self: Self, buffer: bytes = bytes()) -> None:
        """
        Construct a new SerializedType.

        Args:
            buffer: The buffer containing the data for the SerializedType.
        """
        self.buffer = buffer

    @classmethod
    @abstractmethod
    def from_parser(  # noqa: D102
        cls: Type[Self],
        parser: BinaryParser,
        # length_hint is Any so that subclasses can choose whether or not to require it.
        length_hint: Any,  # noqa: ANN401
    ) -> Self:
        pass

    @classmethod
    @abstractmethod
    def from_value(cls: Type[Self], value: Any) -> Self:  # noqa: D102, ANN401
        # Note: Its difficult to specify a concrete type for value parameter because
        # child-classes accept a variety of types.
        pass

    def to_byte_sink(self: Self, bytesink: bytearray) -> None:
        """
        Write the bytes representation of a SerializedType to a bytearray.

        Args:
            bytesink: The bytearray to write self.buffer to.

        Returns: None
        """
        bytesink.extend(self.buffer)

    def __bytes__(self: Self) -> bytes:
        """
        Get the bytes representation of a SerializedType.

        Returns:
            The bytes representation of the SerializedType.
        """
        return self.buffer

    def to_json(self: Self) -> Any:  # noqa: ANN401
        """
        Returns the JSON representation of a SerializedType.

        If not overridden, returns hex string representation of bytes.

        Returns:
            The JSON representation of the SerializedType.
        """
        return self.to_hex()

    def __str__(self: Self) -> str:
        """
        Returns the hex string representation of self.buffer.

        Returns:
            The hex string representation of self.buffer.
        """
        return self.to_hex()

    def to_hex(self: Self) -> str:
        """
        Get the hex representation of a SerializedType's bytes.

        Returns:
            The hex string representation of the SerializedType's bytes.
        """
        return self.buffer.hex().upper()

    def __len__(self: Self) -> int:
        """
        Get the length of a SerializedType's bytes.

        Returns:
            The number of bytes.
        """
        return len(self.buffer)



"""Collection of public constants for XRPL."""

import re
from decimal import Context
from enum import Enum
from typing import Pattern

from typing_extensions import Final


class CryptoAlgorithm(str, Enum):
    """Represents the supported cryptography algorithms."""

    ED25519 = "ed25519"
    SECP256K1 = "secp256k1"


class XRPLException(Exception):
    """Base Exception for XRPL library."""

    pass


SPECIAL_CHARS_CURRENCY_CODE = re.escape("?!@#$%^&*(){}[]<>|")
ISO_CURRENCY_REGEX: Final[Pattern[str]] = re.compile(
    "[A-Za-z0-9" + SPECIAL_CHARS_CURRENCY_CODE + "]{3}"
)
"""
Matches ISO currencies like "USD" or "EUR" in the format allowed by XRPL.
Check the docs for more information:
https://xrpl.org/currency-formats.html#standard-currency-codes

:meta private:
"""

HEX_REGEX: Final[Pattern[str]] = re.compile(r"^[0-9A-Fa-f]+$")

HEX_CURRENCY_REGEX: Final[Pattern[str]] = re.compile("[A-F0-9]{40}")
"""
Matches hex-encoded currencies in the format allowed by XRPL.

:meta private:
"""

HEX_MPTID_REGEX: Final[Pattern[str]] = re.compile(r"^[0-9A-Fa-f]{48}$")

# Constants for validating amounts.
MIN_IOU_EXPONENT: Final[int] = -96
"""
:meta private:
"""
MAX_IOU_EXPONENT: Final[int] = 80
"""
:meta private:
"""
MAX_IOU_PRECISION: Final[int] = 16
"""
:meta private:
"""
MIN_IOU_MANTISSA: Final[int] = 10**15
"""
:meta private:
"""
MAX_IOU_MANTISSA: Final[int] = 10**16 - 1
"""
:meta private:
"""

# Configure Decimal
IOU_DECIMAL_CONTEXT: Final[Context] = Context(
    prec=MAX_IOU_PRECISION, Emax=MAX_IOU_EXPONENT, Emin=MIN_IOU_EXPONENT
)
"""
Decimal context for working with IOUs.
:meta private:
"""


DROPS_DECIMAL_CONTEXT: Final[Context] = Context(prec=18, Emin=0, Emax=18)
"""
Decimal context for working with drops.
:meta private:
"""



"""Model for AMMDelete transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field

from xrpl.models.currencies import Currency
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class AMMDelete(Transaction):
    """
    Delete an empty Automated Market Maker (AMM) instance that could not be fully
    deleted automatically.

    Tip: The AMMWithdraw transaction automatically tries to delete an AMM, along with
    associated ledger entries such as empty trust lines, if it withdrew all the assets
    from the AMM's pool. However, if there are too many trust lines to the AMM account
    to remove in one transaction, it may stop before fully removing the AMM. Similarly,
    an AMMDelete transaction removes up to a maximum number of trust lines; in extreme
    cases, it may take several AMMDelete transactions to fully delete the trust lines
    and the associated AMM. In all cases, the AMM ledger entry and AMM account are
    deleted by the last such transaction.
    """

    asset: Currency = REQUIRED  # type: ignore
    """
    The definition for one of the assets in the AMM's pool. This field is required.
    """

    asset2: Currency = REQUIRED  # type: ignore
    """
    The definition for the other asset in the AMM's pool. This field is required.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.AMM_DELETE,
        init=False,
    )



"""This request gets information about a network's amendments."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional

from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class Feature(Request):
    """The `feature` method gets information about a network's amendments."""

    feature: Optional[str] = None
    """
    The hex-encoded feature hash.
    """

    method: RequestMethod = field(default=RequestMethod.FEATURE, init=False)
